# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from threading import local

_local = local()


class Gate(object):
    def __init__(self):
        self.local = local()

    def setup(self, *args, **kwargs):
        if hasattr(self.local, 'gate_registered_names'):
            raise RuntimeError('Gate already started.')

        self.local.gate_registered_names = set()

    def teardown(self):
        if not hasattr(self.local, 'gate_registered_names'):
            raise RuntimeError('Gate not started.')

        for name in self.local.gate_registered_names:
            delattr(self.local, name)
        delattr(self.local, 'gate_registered_names')

    def __call__(self, *args, **kwargs):
        self.setup(*args, **kwargs)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.teardown()

    def attach(self, name, value):
        self.local.gate_registered_names.add(name)
        setattr(self.local, name, value)

    def contains(self, name):
        return name in self.local.gate_registered_names

    def retrieve(self, name):
        if not self.contains(name):
            raise ValueError('Property `{:s}` is not attached.'.format(
                name))
        return getattr(self.local, name)

    def release(self, name):
        self.local.gate_registered_names.remove(name)
        delattr(self.local, name)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .utils import FileStorage
from decimal import Decimal
from jinja2.runtime import Undefined
from jinja2.exceptions import TemplateNotFound
from os import path
from pandora.compat import PY2
from pandora.compat import text_type
from time import mktime
from time import time
from werkzeug.datastructures import Headers
from werkzeug.http import dump_cookie
from werkzeug.http import http_date
from werkzeug.http import HTTP_STATUS_CODES
from werkzeug.http import is_resource_modified
from werkzeug.wsgi import wrap_file
from zlib import adler32
import datetime as dt
import jinja2
import json
import mimetypes


class Response(object):
    def __init__(self, body='', headers=None, status_code=200):
        self.status_code = status_code
        self.headers = Headers(headers)
        self.body = body

    def get_status(self):
        return '{:d} {:s}'.format(self.status_code,
                                  HTTP_STATUS_CODES[self.status_code])

    def get_body(self):
        return self.body

    def prepare_response(self, request=None):
        if PY2:
            return [self.get_body().encode('utf-8')]
        else:
            return [bytes(self.get_body(), 'UTF8')]

    def respond(self, request):
        return (self.get_status(), self.headers.to_wsgi_list(),
                self.prepare_response(request))

    def __call__(self, request, start_response):
        status, headers, response = self.respond(request)
        start_response(status, headers)
        return response

    def set_cookie(self, name, value='', max_age=None,
                   path='/', domain=None, secure=False, httponly=True):
        self.headers.add('Set-Cookie', dump_cookie(
            name, value=str(value), max_age=max_age,
            path=path, domain=domain,
            secure=secure, httponly=httponly,
            charset='utf-8', sync_expires=True))

    def del_cookie(self, name, path='/', domain=None, secure=False,
                   httponly=True):
        self.set_cookie(name, max_age=0, path=path, domain=domain,
                        secure=secure, httponly=httponly)


class JsonResponse(Response):
    def __init__(self, data, headers=None, status_code=200):
        self.data = data
        super(JsonResponse, self).__init__(headers=headers,
                                           status_code=status_code)
        self.headers.set('Content-Type', 'application/json')

    def get_body(self):
        return json.dumps(self.data, default=self.json_default)

    def json_default(self, value):
        if type(value) is Undefined:
            return None
        elif isinstance(value, (dt.datetime, dt.date)):
            return value.isoformat()
        elif isinstance(value, Decimal):
            return str(value)
        elif isinstance(value, FileStorage):
            return value.filename
        elif hasattr(value, 'to_dict'):
            return value.to_dict()
        elif hasattr(value, '__iter__'):
            return [v for v in value]
        elif type(value).__str__ is not object.__str__:
            return text_type(value)
        else:
            return repr(value)


class HtmlResponse(Response):
    def __init__(self, body, headers=None, status_code=200):
        super(HtmlResponse, self).__init__(body, headers, status_code)
        self.headers.set('Content-Type', 'text/html')


class TemplateResponse(HtmlResponse):
    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(path.abspath('.')))

    def __init__(self, template, context, headers=None, status_code=200):
        self.template = self.jinja.get_template(template)
        self.context = context

        super(HtmlResponse, self).__init__(headers=headers,
                                           status_code=status_code)

    def get_body(self):
        return self.template.render(**self.context)


class WrappedTemplateResponse(TemplateResponse):
    BASE_TEMPLATE = 'base.html'

    def __init__(self, template, context, headers=None, status_code=200,
                 base_template=None):
        self.base_template = self.jinja.get_template(
            base_template or self.BASE_TEMPLATE)

        super(WrappedTemplateResponse, self).__init__(template, context,
                                                      headers, status_code)

    def get_body(self):
        return self.base_template.render({
            'body': self.template.render(self.context),
            'context': self.context,
        })


class FileResponse(Response):
    def __init__(self, filename, as_attachment=False,
                 cache_timeout=60 * 60 * 12, headers=None, status_code=200):
        if not path.exists(filename):
            raise ValueError('File `{:s}` does not exist.'.format(
                filename))

        super(FileResponse, self).__init__(headers=headers,
                                           status_code=status_code)
        self.filename = filename
        self.cache_timeout = cache_timeout
        self.as_attachment = bool(as_attachment)

        if as_attachment is True:
            self.attachment_filename = path.basename(self.filename)
        elif as_attachment:
            self.attachment_filename = as_attachment
        else:
            self.attachment_filename = None

    @property
    def size(self):
        return path.getsize(self.filename)

    @property
    def mtime(self):
        return dt.datetime.fromtimestamp(int(path.getmtime(self.filename)))

    @property
    def etag(self):
        if PY2:
            filename = self.filename
        else:
            filename = bytes(self.filename, 'UTF8')

        return 'etag-%d-%s-%s' % (
            mktime(self.mtime.timetuple()),
            self.size,
            adler32(filename) & 0xffffffff
        )

    def prepare_response(self, request=None):
        if request is None:
            environ = dict()
        else:
            environ = request.environ

        guessed_type = mimetypes.guess_type(self.filename)
        mime_type = guessed_type[0] or 'text/plain'

        stream = open(self.filename)

        self.headers['Date'] = http_date()

        if self.cache_timeout and request:
            self.headers.extend((
                ('Etag', '"{:s}"'.format(self.etag)),
                ('Cache-Control', 'max-age={:d}, public'.format(
                    self.cache_timeout)),
            ))

            if not is_resource_modified(environ, self.etag):
                stream.close()
                self.status_code = 304
                return []

            self.headers['Expires'] = http_date(time() + self.cache_timeout)
        else:
            self.headers['Cache-Control'] = 'public'

        self.headers.extend((
            ('Content-Type', mime_type),
            ('Content-Length', str(self.size)),
            ('Last-Modified', http_date(self.mtime))
        ))

        if self.as_attachment:
            self.headers.set('Content-Disposition', 'attachment',
                             filename=self.attachment_filename)

        return wrap_file(environ, stream)


def template_or_json_response(request, template, context, headers=None,
                              status_code=200, template_response_class=None,
                              json_response_class=None):
    if template_response_class is None:
        template_response_class = TemplateResponse
    if json_response_class is None:
        json_response_class = JsonResponse

    trc = template_response_class
    jrc = json_response_class

    if request.want_json:
        return jrc(context, headers, status_code)
    else:
        try:
            return trc(template, context, headers, status_code)
        except TemplateNotFound:
            return jrc(context, headers, status_code)


def wrapped_template_or_json_response(request, template, context, headers=None,
                                      status_code=200,
                                      template_response_class=None,
                                      json_response_class=None):
    if template_response_class is None:
        template_response_class = WrappedTemplateResponse
    if json_response_class is None:
        json_response_class = JsonResponse

    return template_or_json_response(request, template, context, headers,
                                     status_code, template_response_class,
                                     json_response_class)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .utils import FileStorage
from .utils import flatten_multidict
from werkzeug.formparser import parse_form_data
from werkzeug.http import parse_cookie
from werkzeug.urls import url_decode
import json


class Request(object):
    def __init__(self, environ):
        self.environ = environ

        self.host = environ['HTTP_HOST']
        self.path = environ['PATH_INFO']
        self.method = environ['REQUEST_METHOD']

        self.ip_address = environ.get('REMOTE_ADDR', None)

        self.accept_type = environ.get('HTTP_ACCEPT', '')
        self.content_type = environ.get('CONTENT_TYPE', '')

        self.want_json = 'application/json' in self.accept_type
        self.want_html = 'text/html' in self.accept_type

        stream, form, files = parse_form_data(environ)

        self.query = flatten_multidict(url_decode(environ['QUERY_STRING']))
        self.form = flatten_multidict(form)
        self.files = flatten_multidict(files, lambda v: len(v.filename),
                                       FileStorage.from_original)

        self.cookies = parse_cookie(environ)

        if 'application/json' in self.content_type:
            try:
                self.json = json.loads(stream.read().decode('utf-8'))
            except ValueError:
                self.json = None
        else:
            self.json = None

        self.params = dict()
        self.params.update(self.query)
        self.params.update(self.form)
        self.params.update(self.files)
        if self.json:
            self.params.update(self.json)

    def as_dict(self):
        return {
            'host': self.host,
            'path': self.path,
            'method': self.method,

            'ip_address': self.ip_address,

            'accept_type': self.accept_type,
            'content_type': self.content_type,

            'want_json': self.want_json,
            'want_html': self.want_html,

            'query': self.query,
            'form': self.form,
            'files': self.files,
            'cookies': self.cookies,
            'json': self.json,

            'params': self.params,
        }

    def get_cookie(self, name, default=None):
        return self.cookies.get(name, default)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from os import path
from os import SEEK_END
from os import system
from os import unlink
from psutil import NoSuchProcess
from psutil import Process
import click
import curses
import sys

try:
    import gunicorn

    assert gunicorn
except ImportError:
    raise ValueError('Poort::cli requires Gunicorn to run!')

try:
    sys.path.append('.')
    import config
except ImportError:
    raise ValueError('Could not import `config.py`.')


@click.group()
@click.option('--debug/--no-debug', default=False)
def main(debug):
    if debug:
        click.echo('proc_name = %s' % config.proc_name)
        click.echo('bind      = %s' % config.bind)
        click.echo('pidfile   = %s' % config.pidfile)


def get_process():
    if not path.exists(config.pidfile):
        return False

    try:
        with open(config.pidfile) as stream:
            return Process(int(stream.read().strip()))
    except NoSuchProcess:
        unlink(config.pidfile)
        return False


@main.command('start')
def start():
    if get_process() is not False:
        click.secho('Application is already running.', err=True, fg='red')
        raise click.Abort

    click.secho('Starting your application.', fg='green')
    system('gunicorn -c config.py website.app:application')


@main.command('stop')
@click.option('--graceful/--quick', default=True)
def stop(graceful):
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    if graceful:
        click.secho('Stopping your application.', fg='green')
        system('kill -TERM `cat %s`' % config.pidfile)
    else:
        click.secho('Stopping your application (force-quit).', fg='purple')
        system('kill -QUIT `cat %s`' % config.pidfile)


@main.command('reload')
def reload():
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    click.secho('Restarting application.', fg='green')
    system('kill -HUP `cat %s`' % config.pidfile)


@main.command('scale')
@click.argument('way', type=click.Choice(['up', 'down']))
@click.argument('amount', default=1)
def scale(way, amount):
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    if amount == 0:
        click.secho('Cannot scale 0.', err=True, fg='red')
        raise click.Abort

    if way == 'down':
        click.secho('Scaling application %d down.' % amount, fg='green')
        for i in range(amount):
            system('kill -TTOU `cat %s`' % config.pidfile)
    elif way == 'up':
        click.secho('Scaling application %d up.' % amount, fg='green')
        for i in range(amount):
            system('kill -TTIN `cat %s`' % config.pidfile)


@main.command('status')
@click.option('--watch/--once', default=True)
@click.option('--delay', default=1)
def status(watch, delay):
    if get_process() is False:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    process = get_process()
    if not process:
        click.secho('Application is not running.', err=True, fg='red')
        raise click.Abort

    screen = curses.initscr()
    curses.halfdelay(delay * 10)
    curses.noecho()

    def line(name, proc, y, x=2):
        cpu_percentage = proc.cpu_percent(None)
        memory = proc.memory_info()

        screen.addstr(y, x, '%-10s %5.d   %4.1f%%   %5.1fM' % (
            name, proc.pid,
            cpu_percentage, memory.rss / 1024 / 1024))

    status = 'Waiting...'

    running = True
    while running:
        try:
            children = process.children()
            workers = len(children)

            screen.erase()
            screen.addstr(1, 2, config.proc_name or 'Unnamed')
            screen.addstr(4, 2, 'Running with %d workers (default is %d)' % (
                workers, config.workers))

            screen.addstr(6, 2, '%-10s %5s   %5s   %6s' % (
                'Name', 'PID', 'CPU', 'Mem'))
            screen.addstr(7, 2, '-' * 40)

            line('Master', process, 8)

            for cx, child in enumerate(children):
                line('Worker %d' % (cx + 1), child, 9 + cx)

            usage = '(q)uit (u)pscale (d)ownscale (r)estart'

            screen.addstr(9 + workers + 2, 2, '%-20s -- %s' % (
                status, usage))

            height = screen.getmaxyx()[0] - 1

            top = 9 + workers + 5
            max_lines = height - top

            with open(config.errorlog) as stream:
                contents = tail(stream, max_lines)

            for lx, content in enumerate(contents):
                screen.addstr(top + lx, 2, content)

            char = screen.getch()
            if char != curses.ERR:
                key = chr(char)

                if key == 'q':
                    status = 'Quit'
                    running = False
                elif key == 'u':
                    status = 'Scaling up'
                    system('kill -TTIN `cat %s`' % config.pidfile)
                elif key == 'd':
                    status = 'Scaling down'
                    system('kill -TTOU `cat %s`' % config.pidfile)
                elif key == 'r':
                    status = 'Restarting'
                    system('kill -HUP `cat %s`' % config.pidfile)
                else:
                    status = 'Unknown key'
            else:
                status = 'Waiting...'

        except KeyboardInterrupt:
            running = False
        else:
            if not watch:
                running = False

    curses.endwin()


def tail(f, lines=1, _buffer=4098):
    """Tail a file and get X lines from the end"""
    # place holder for the lines found
    lines_found = []

    # block counter will be multiplied by buffer
    # to get the block size from the end
    block_counter = -1

    # loop until we find X lines
    while len(lines_found) < lines:
        try:
            f.seek(block_counter * _buffer, SEEK_END)
        except IOError:  # too small or too many lines requested
            f.seek(0)
            lines_found = f.readlines()
            break

        lines_found = f.readlines()

        if len(lines_found) > lines:
            break

        block_counter -= 1

    return lines_found[-lines:]
